/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.maven.wagons.aws.s3;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.ClientConfiguration;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3Client;
import com.amazonaws.services.s3.model.CannedAccessControlList;
import com.amazonaws.services.s3.model.GetObjectRequest;
import com.amazonaws.services.s3.model.ListObjectsRequest;
import com.amazonaws.services.s3.model.ObjectListing;
import com.amazonaws.services.s3.model.ObjectMetadata;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectSummary;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.activation.MimetypesFileTypeMap;

import org.apache.commons.io.IOUtils;

import org.apache.maven.wagon.ConnectionException;
import org.apache.maven.wagon.InputData;
import org.apache.maven.wagon.OutputData;
import org.apache.maven.wagon.ResourceDoesNotExistException;
import org.apache.maven.wagon.StreamWagon;
import org.apache.maven.wagon.TransferFailedException;
import org.apache.maven.wagon.authentication.AuthenticationException;
import org.apache.maven.wagon.authentication.AuthenticationInfo;
import org.apache.maven.wagon.authorization.AuthorizationException;
import org.apache.maven.wagon.proxy.ProxyInfo;
import org.apache.maven.wagon.resource.Resource;

/**
 * Amazon S3 wagon provider implementation.
 */
public class S3Wagon extends StreamWagon {

    /**
     * Maps file extensions to mime-types. Will read META-INF/mime.types by default.
     */
    private static final MimetypesFileTypeMap MIME_TYPES = new MimetypesFileTypeMap();

    private AmazonS3 s3client;
    private String bucket;
    private String keyPrefix;

    @Override
    public void closeConnection() throws ConnectionException {
    }

    @Override
    public void fillInputData(InputData in) throws TransferFailedException, ResourceDoesNotExistException, AuthorizationException {
        String key = keyPrefix + in.getResource().getName();
        S3Object object;
        try {
            object = s3client.getObject(bucket, key);
        } catch (AmazonServiceException e) {
            if (e.getStatusCode() == 404) {
                throw new ResourceDoesNotExistException(String.format("key %s does not exist in S3 bucket %s", key, bucket));
            } else if (e.getStatusCode() == 403) {
                throw new ResourceDoesNotExistException(String.format("403 implies that key %s does not exist in bucket %s", key, bucket), e);
            }
            throw new TransferFailedException(String.format("can't get object %s from S4 bucket %s", key, bucket), e);
        }
        in.getResource().setContentLength(object.getObjectMetadata().getContentLength());
        in.getResource().setLastModified(object.getObjectMetadata().getLastModified().getTime());
        in.setInputStream(object.getObjectContent());
    }

    @Override
    public void fillOutputData(OutputData out) throws TransferFailedException {
        throw new UnsupportedOperationException("this call is not supported");
    }

    @Override
    public void get(String resourceName, File destination) throws ResourceDoesNotExistException, TransferFailedException {
        Resource resource = new Resource(resourceName);
        fireGetInitiated(resource, destination);
        String key = keyPrefix + resourceName;
        try {
            fireGetStarted(resource, destination);
            // This is a bit more efficient than copying stream
            s3client.getObject(new GetObjectRequest(bucket, key), destination);
            fireGetCompleted(resource, destination);
        } catch (AmazonServiceException e) {
            if (e.getStatusCode() == 404) {
                throw new ResourceDoesNotExistException(String.format("key %s does not exist in bucket %s", key, bucket), e);
            } else if (e.getStatusCode() == 403) {
                throw new ResourceDoesNotExistException(String.format("403 implies that key %s does not exist in bucket %s", key, bucket), e);
            } else {
                throw new TransferFailedException(String.format("getting metadata of key %s failed", key), e);
            }
        }
    }

    @Override
    public List<String> getFileList(String destinationDirectory) throws TransferFailedException, ResourceDoesNotExistException {
        String path = keyPrefix + destinationDirectory;
        if (!path.endsWith("/")) {
            path += "/";
        }
        // Since S3 does not have concept of directory, result contains all contents with given prefix
        ObjectListing result = s3client.listObjects(new ListObjectsRequest().withBucketName(bucket).withPrefix(path).withDelimiter("/"));
        if (result.getObjectSummaries().isEmpty()) {
            throw new ResourceDoesNotExistException(String.format("no keys exist with prefix %s", path));
        }
        Set<String> results = new HashSet<String>();
        for (S3ObjectSummary summary : result.getObjectSummaries()) {
            String name = summary.getKey().substring(path.length());
            if (name.indexOf('/') == -1) {
                results.add(name);
            } else {
                results.add(name.substring(0, name.indexOf('/')));
            }
        }
        return new ArrayList<String>(results);
    }

    @Override
    public boolean getIfNewer(String resourceName, File destination, long timestamp) throws ResourceDoesNotExistException, TransferFailedException {
        boolean result = true;
        ObjectMetadata meta = getRequiredMetadata(resourceName);
        if (meta == null) {
            result = false;
        } else {
            get(resourceName, destination);
        }
        return result;
    }

    @Override
    public boolean getIfNewerToStream(String resourceName, OutputStream out, long timestamp) throws ResourceDoesNotExistException, TransferFailedException {
        boolean result = true;
        ObjectMetadata meta = getRequiredMetadata(resourceName);
        if (meta == null) {
            result = false;
        } else {
            Resource resource = new Resource(resourceName);
            fireGetInitiated(resource, null);
            InputStream in = s3client.getObject(bucket, keyPrefix).getObjectContent();
            try {
                fireGetStarted(resource, null);
                IOUtils.copy(in, out);
                out.flush();
                out.close();
                fireGetCompleted(resource, null);
            } catch (IOException e) {
                throw new TransferFailedException("Stream copy failed", e);
            } finally {
                IOUtils.closeQuietly(in);
            }
        }
        return result;
    }

    @Override
    protected void openConnectionInternal() throws ConnectionException, AuthenticationException {
        AuthenticationInfo auth = getAuthenticationInfo();
        if (auth == null) {
            throw new AuthenticationException("S3 access requires authentication information");
        }
        if (auth.getUserName() == null || auth.getUserName().trim().length() < 1) {
            throw new AuthenticationException("Tag <username> must set to valid AWS access key ID in server configuration, either in pom.xml or settings.xml");
        }
        if (auth.getPassword() == null || auth.getPassword().trim().length() < 1) {
            throw new AuthenticationException("Tag <password> must set to valid AWS secret key in server configuration, either in pom.xml or settings.xml");
        }
        AWSCredentials credentials = new BasicAWSCredentials(auth.getUserName(), auth.getPassword());
        ClientConfiguration config = new ClientConfiguration();
        config.setConnectionTimeout(getTimeout());
        config.setSocketTimeout(getTimeout());
        ProxyInfo proxy = getProxyInfo();
        if (proxy != null) {
            config.setProxyDomain(proxy.getNtlmDomain());
            config.setProxyHost(proxy.getHost());
            config.setProxyPassword(proxy.getPassword());
            config.setProxyPort(proxy.getPort());
            config.setProxyUsername(proxy.getUserName());
            config.setProxyWorkstation(proxy.getNtlmHost());
        }
        s3client = new AmazonS3Client(credentials, config);
        bucket = getRepository().getHost();
        String prefix = null;
        if (getRepository().getBasedir() != null) {
            prefix = getRepository().getBasedir().trim();
        } else {
            prefix = "";
        }
        if (!prefix.endsWith("/")) {
            prefix += "/";
        }
        if (prefix.startsWith("/")) {
            prefix = prefix.substring(1);
        }
        keyPrefix = prefix;
    }

    @Override
    public void put(File source, String destination) throws TransferFailedException, ResourceDoesNotExistException {
        try {
            putFromStream(new FileInputStream(source), source, destination, source.length(), source.lastModified());
        } catch (FileNotFoundException e) {
            throw new ResourceDoesNotExistException("source file " + source + " does not exist", e);
        }
    }

    @Override
    public void putDirectory(File sourceDirectory, String destinationDirectory) throws TransferFailedException, ResourceDoesNotExistException, AuthorizationException {
        String destDir = destinationDirectory;
        if (destDir.endsWith("/")) {
            destDir = destDir.substring(0, destDir.length() - 1);
        }
        if (destDir.equals(".")) {
            destDir = "";
        }
        for (File file : sourceDirectory.listFiles()) {
            String dest = null;
            if (destDir == null || destDir.trim().length() < 1) {
                dest = file.getName();
            } else {
                dest = (destDir + "/" + file.getName());
            }
            if (file.isDirectory()) {
                putDirectory(file, dest);
            } else {
                put(file, dest);
            }
        }
    }

    @Override
    public void putFromStream(InputStream in, String destination) throws TransferFailedException, ResourceDoesNotExistException {
        putFromStream(in, null, destination, -1, System.currentTimeMillis());
    }

    @Override
    public void putFromStream(InputStream in, String destination, long contentLength, long lastModified) throws TransferFailedException, ResourceDoesNotExistException {
        putFromStream(in, null, destination, contentLength, lastModified);
    }

    @Override
    public boolean resourceExists(String resourceName) throws TransferFailedException, AuthorizationException {
        boolean result = true;
        String key = keyPrefix + resourceName;
        try {
            s3client.getObjectMetadata(bucket, key);
        } catch (AmazonServiceException e) {
            if (e.getStatusCode() == 404) {
                result = false;
            } else {
                throw new TransferFailedException(String.format("resource %s cannot be verified", key), e);
            }
        }
        return result;
    }

    @Override
    public boolean supportsDirectoryCopy() {
        return true;
    }

    private void putFromStream(InputStream in, File inFile, String destination, long contentLength, long lastModified) {
        Resource resource = new Resource(destination);
        firePutInitiated(resource, inFile);
        String key = keyPrefix + destination;
        key = resolveCanonicalKey(key);
        ObjectMetadata meta = new ObjectMetadata();
        if (contentLength != -1) {
            meta.setContentLength(contentLength);
        }
        meta.setLastModified(new Date(lastModified));
        String mimeType = MIME_TYPES.getContentType(destination);
        if (mimeType != null) {
            meta.setContentType(mimeType);
        } else {
            fireTransferDebug(String.format("mime-type of %s cannot be determined", destination));
        }
        try {
            firePutStarted(resource, inFile);
            s3client.putObject(bucket, key, in, meta);
            s3client.setObjectAcl(bucket, key, CannedAccessControlList.PublicRead);
            firePutCompleted(resource, inFile);
        } finally {
            IOUtils.closeQuietly(in);
        }
    }

    private ObjectMetadata getRequiredMetadata(String resourceName) throws ResourceDoesNotExistException, TransferFailedException {
        String key = keyPrefix + resourceName;
        try {
            return s3client.getObjectMetadata(bucket, key);
        } catch (AmazonServiceException e) {
            if (e.getStatusCode() == 404) {
                throw new ResourceDoesNotExistException(String.format("key %s does not exist in bucket %s", key, bucket), e);
            } else {
                throw new TransferFailedException(String.format("getting metadata for key %s failed", key), e);
            }
        }
    }

    /**
     * Maven site plugin uses unix notation for relative directory references, like '.. and '.', however since S3 is
     * not a proper directory/file based system, it does not support this notation and so this method fixes any of these
     * references.
     */
    private String resolveCanonicalKey(String key) {
//        fireTransferDebug(String.format("resolving: %s", key));
        String result = '/' + key;
        // find all '..' references and eliminate
        int index = result.indexOf("../");
        while (index >= 0) {
//            fireTransferDebug(String.format(".. reference is at: %s", index));
            int parentDirectoryIndex = result.lastIndexOf("/", index);
//            fireTransferDebug(String.format("first parent directory index: %s", parentDirectoryIndex));
            parentDirectoryIndex = result.lastIndexOf("/", parentDirectoryIndex - 1);
//            fireTransferDebug(String.format("second parent directory index: %s", parentDirectoryIndex));
            result = result.substring(0, parentDirectoryIndex + 1) + result.substring(index + 3);
//            fireTransferDebug(String.format("result is now: %s", result));
            index = result.indexOf("..");
        }
        // find all '.' references and eliminate
        index = result.indexOf("./");
        while (index >= 0) {
            result = result.substring(0, index - 1) + result.substring(index + 1);
            index = result.indexOf("./");
        }
        if (result.startsWith("/")) {
            result = result.substring(1);
        }
//        fireTransferDebug(String.format("resolved to: %s", result));
        return result;
    }

}
